'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = expand;

var _value = require('./value');

var _pair = require('./value/pair');

var _pair2 = _interopRequireDefault(_pair);

var _syntaxRules = require('./transformer/syntaxRules');

var _syntaxRules2 = _interopRequireDefault(_syntaxRules);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Statically expands the AST using macros... This is equvalent to
// macro-expansion time (specified in R6RS)

var transformers = {
  'syntax-rules': function syntaxRules(code) {
    return new _syntaxRules2.default(code);
  }
};

function unquoteHandler(frame, code, stack) {
  // Create new stack, allowing expand requests again
  return new _pair2.default({
    code: code,
    scope: frame.scope,
    ignore: false
  }, stack);
}

function letSyntax(frame, code, stack) {
  var name = code.cdr.car;
  if (!name || name.type !== _value.SYMBOL) {
    throw new Error('let-syntax expects a symbol');
  }
  if (stack.scope.car[name.value]) {
    throw new Error('Syntax ' + name.value + ' conflicts');
  }
  var transCode = code.cdr.cdr.car;
  var transName = transCode.car;
  if (!transName || transName.type !== _value.SYMBOL) {
    throw new Error('Transformer name must be a symbol');
  }
  var transFunc = transformers[transName.value];
  if (!transFunc) {
    throw new Error('Unsupported transformer type');
  }
  var transObj = transFunc(transCode);
  // Apply the transform object to root scope
  stack.scope.car[name.value] = transObj;
  // Don't process child nodes
  frame.result = code;
  return stack;
}

var basicKeywords = {
  'lambda': function lambda(frame, code, stack) {
    // Create new scope and stack
    return new _pair2.default({
      code: code,
      scope: new _pair2.default({}, frame.scope)
    }, stack);
  },
  'define-syntax': function defineSyntax(frame, code, stack, rootScope) {
    // Create the transformer object
    var name = code.cdr.car;
    if (!name || name.type !== _value.SYMBOL) {
      throw new Error('define-syntax expects a symbol');
    }
    if (rootScope[name.value]) {
      throw new Error('Syntax ' + name.value + ' conflicts');
    }
    var transCode = code.cdr.cdr.car;
    var transName = transCode.car;
    if (!transName || transName.type !== _value.SYMBOL) {
      throw new Error('Transformer name must be a symbol');
    }
    var transFunc = transformers[transName.value];
    if (!transFunc) {
      throw new Error('Unsupported transformer type');
    }
    var transObj = transFunc(transCode);
    // Apply the transform object to root scope
    rootScope[name.value] = transObj;
    // Don't process child nodes
    frame.result = code;
    return stack;
  },
  'let-syntax': letSyntax,
  'letrec-syntax': letSyntax,
  'quote': function quote(frame, code, stack) {
    // Don't process child nodes - it shouldn't be processed!
    frame.result = code;
    return stack;
  },
  'quasiquote': function quasiquote(frame, code, stack) {
    // Create new stack, ignoring any expand requests
    return new _pair2.default({
      code: code,
      scope: frame.scope,
      ignore: true
    }, stack);
  },
  'unquote': unquoteHandler,
  'unquote-splicing': unquoteHandler
};

function findScope(scope, name) {
  var scopeNode = scope;
  while (scopeNode != null) {
    if (scopeNode.car[name]) {
      // Found it!
      return scopeNode.car[name];
    }
    scopeNode = scopeNode.cdr;
  }
}

// Root scope stores the 'define-syntax' status.
function expand(code) {
  var rootScope = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
  var upperScope = arguments[2];

  // Well, we should visit whole node and perform transformation if required.
  // Traverse ALL the node!!
  var stack = new _pair2.default({
    code: code,
    scope: new _pair2.default(rootScope, upperScope)
  });
  while (stack != null) {
    var frame = stack.car;
    var _code = frame.code;
    if (frame.code.type === _value.PAIR) _code = frame.code.car;
    // Process the node...
    if (_code && _code.type === _value.PAIR) {
      // Traverse child node! ... but before doing that, we should determine
      // if the pair is subject to expand.
      // Or, we should check if it creates new scope (lambda), or it defines
      // new syntax. (define-syntax, let-syntax, etc)
      if (frame.result == null) {
        if (_code.car && _code.car.type === _value.SYMBOL) {
          var keywordHandler = basicKeywords[_code.car.value];
          if (keywordHandler) {
            stack = keywordHandler(frame, _code, stack, rootScope);
            continue;
          }
        }
        // Run transform until no transformer is available.
        while (_code.car && _code.car.type === _value.SYMBOL && !frame.ignore) {
          // Traverse scope information, and find the symbol
          var transformer = findScope(frame.scope, _code.car.value);
          if (transformer) {
            var original = _code;
            _code = transformer.exec(original);
            if (original === _code) break;
          } else {
            break;
          }
        }
        if (_code.type === _value.PAIR) {
          // Create child stack frame.
          stack = new _pair2.default({
            code: _code,
            scope: frame.scope,
            ignore: frame.ignore
          }, stack);
        } else {
          // If transformer has transformed the code into other types,
          // we need to handle it like any other normal value.
          frame.code = new _pair2.default(_code);
        }
        continue;
      } else {
        _code = frame.result;
        frame.result = null;
      }
    } else if (_code && _code.type === _value.SYMBOL) {}
    // Is symbol itself subject to transformation? I suppose not.

    // Copy current node to the output.
    if (frame.code.type === _value.PAIR) {
      var newPair = new _pair2.default(_code, null);
      if (frame.outputTail) {
        frame.outputTail.cdr = newPair;
        frame.outputTail = newPair;
      } else {
        frame.outputHead = newPair;
        frame.outputTail = newPair;
      }
    } else {
      // Handle CDR value differently...
      if (frame.outputTail) {
        frame.outputTail.cdr = _code;
      } else {
        var _newPair = new _pair2.default(null, _code);
        frame.outputHead = _newPair;
        frame.outputTail = _newPair;
      }
    }
    // Continue to next stack entry.
    if (frame.code.type === _value.PAIR) {
      frame.code = frame.code.cdr;
    } else {
      frame.code = null;
    }
    if (frame.code == null) {
      // Done! Pop current stack..
      stack = stack.cdr;
      if (stack != null) {
        stack.car.result = frame.outputHead;
      } else {
        // Traversal finished. return current output...
        return frame.outputHead;
      }
    }
  }
}