'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _wrapSymbols;

exports.default = parse;

var _boolean = require('./value/boolean');

var _boolean2 = _interopRequireDefault(_boolean);

var _character = require('./value/character');

var _character2 = _interopRequireDefault(_character);

var _number = require('./value/number');

var _number2 = _interopRequireDefault(_number);

var _pair = require('./value/pair');

var _pair2 = _interopRequireDefault(_pair);

var _string = require('./value/string');

var _string2 = _interopRequireDefault(_string);

var _symbol = require('./value/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

var _tokenizer = require('./tokenizer');

var TOKENS = _interopRequireWildcard(_tokenizer);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var LIST = Symbol('list');

var ROOT = Symbol('root');

function createStackEntry(type, bracketType, token) {
  return {
    head: null,
    tail: null,
    type: type, // Only list is supported for now..
    bracketType: bracketType,
    comment: false, // Should next keyword be ignored?
    special: null, // QUOTE, QUASIQUOTE, etc
    con: 0, // Is . specified? 0 - No, 1 - Yes, it is, 2 - Yes, it did
    line: token && token.line,
    column: token && token.column
  };
}

var wrapSymbols = (_wrapSymbols = {}, _defineProperty(_wrapSymbols, TOKENS.QUOTE, 'quote'), _defineProperty(_wrapSymbols, TOKENS.QUASIQUOTE, 'quasiquote'), _defineProperty(_wrapSymbols, TOKENS.UNQUOTE, 'unquote'), _defineProperty(_wrapSymbols, TOKENS.UNQUOTE_SPLICING, 'unquote-splicing'), _defineProperty(_wrapSymbols, TOKENS.SYNTAX, 'syntax'), _defineProperty(_wrapSymbols, TOKENS.QUASISYNTAX, 'quasisyntax'), _defineProperty(_wrapSymbols, TOKENS.UNSYNTAX, 'unsyntax'), _defineProperty(_wrapSymbols, TOKENS.UNSYNTAX_SPLICING, 'unsyntax-splicing'), _wrapSymbols);

// Wrap the data if special flag is on. COMMENT_IGNORE should be handled
// differently, though.
function wrapData(entry, data, token) {
  var symbol = wrapSymbols[entry.special];
  if (!symbol) return data;
  var symbolVal = new _symbol2.default(symbol);
  if (token) {
    symbolVal.line = token.line;
    symbolVal.column = token.column;
  }
  var result = new _pair2.default(symbolVal, new _pair2.default(data));
  entry.special = null;
  return result;
}

// Pushes the data into the stack entry. Note that this doesn't push the data
// into the stack..
function pushData(entry, data, token) {
  if (entry.con === 2) {
    throw injectError(new Error('Finished pair cannot have more values'), token);
  }
  if (token) {
    data.line = token.line;
    data.column = token.column;
  }
  var wrappedData = wrapData(entry, data, token);
  if (entry.comment) {
    entry.comment = false;
    return;
  }
  if (entry.con === 1) {
    if (entry.tail == null) {
      // Welp, if tail is missing, head is missing too (of course)
      // I don't think this will ever be called..
      entry.head = entry.tail = new _pair2.default(null, wrappedData);
    } else {
      // If not, just attach the value.
      entry.tail.cdr = wrappedData;
    }
    entry.con = 2;
  } else {
    var pair = new _pair2.default(wrappedData, null);
    if (entry.tail == null) {
      // Welp, if tail is missing, head is missing too (of course)
      entry.head = entry.tail = pair;
    } else {
      // If not, just attach the value.
      entry.tail.cdr = pair;
      entry.tail = pair;
    }
  }
  if (token) {
    entry.head.line = token.line;
    entry.head.column = token.column;
  }
}

function injectError(error, token) {
  if (token) {
    error.line = token.line;
    error.column = token.column;
  }
  return error;
}

// Accepts tokens produced by tokenizer, performs syntax analysis and
// returns an AST (S-expression) processable by the interpreter.
function parse(tokens) {
  var stack = [createStackEntry(ROOT)];
  for (var i = 0; i < tokens.length; ++i) {
    var stackEntry = stack[stack.length - 1];
    var token = tokens[i];
    switch (token.type) {
      case TOKENS.LIST_START:
        stack.push(createStackEntry(LIST, token.value, token));
        break;
      case TOKENS.LIST_END:
        // Finalize the current stack..
        if (stack.length <= 1) {
          throw injectError(new Error('Unexpected list closing paren'), token);
        }
        if (stackEntry.con === 1) {
          throw injectError(new Error('List cannot be closed right after cdr specifier'), token);
        }
        if (stackEntry.bracketType !== token.value) {
          throw injectError(new Error('Bracket type does not match'), token);
        }
        // Pop the stack, and refer parent stack.
        stack.pop();
        pushData(stack[stack.length - 1], stackEntry.head || new _pair2.default(), stackEntry);
        break;
      case TOKENS.LIST_CON:
        if (stackEntry.con !== 0) {
          throw injectError(new Error('There can be only one cdr specifier in single list'), token);
        }
        stackEntry.con = 1;
        break;
      case TOKENS.IDENTIFIER:
        pushData(stackEntry, new _symbol2.default(token.value), token);
        break;
      case TOKENS.STRING:
        pushData(stackEntry, new _string2.default(token.value), token);
        break;
      case TOKENS.NUMBER:
        pushData(stackEntry, new _number2.default(token.value), token);
        break;
      case TOKENS.BOOLEAN:
        pushData(stackEntry, new _boolean2.default(token.value), token);
        break;
      case TOKENS.CHARACTER:
        pushData(stackEntry, new _character2.default(token.value), token);
        break;
      case TOKENS.COMMENT_IGNORE:
        stackEntry.comment = true;
        break;
      case TOKENS.ABBERVIATION:
        if (stackEntry.special !== null) {
          throw injectError(new Error('Abberviation should not come twice'), token);
        }
        stackEntry.special = token.value;
        break;
    }
  }
  if (stack.length > 1) {
    throw injectError(new Error('List is not closed'), tokens[tokens.length - 1]);
  }
  return stack[0].head || new _pair2.default();
}