'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = tokenize;
// Lexical analyzer. Converts the code into tokens.

// Lexical syntax complying to R6RS specification. (Actual interpreter may not
// fully implement R6RS.)

var COMMENT = exports.COMMENT = Symbol('comment');
var COMMENT_IGNORE = exports.COMMENT_IGNORE = Symbol('commentIgnore');
var COMMENT_BLOCK_START = exports.COMMENT_BLOCK_START = Symbol('commentBlockStart');
var COMMENT_BLOCK_END = exports.COMMENT_BLOCK_END = Symbol('commentBlockEnd');
var IDENTIFIER = exports.IDENTIFIER = Symbol('identifier');
var NUMBER = exports.NUMBER = Symbol('number');
var BOOLEAN = exports.BOOLEAN = Symbol('boolean');
var CHARACTER = exports.CHARACTER = Symbol('character');
var STRING = exports.STRING = Symbol('string');
var BYTE_VECTOR_START = exports.BYTE_VECTOR_START = Symbol('byteVectorStart');
var VECTOR_START = exports.VECTOR_START = Symbol('vectorStart');
var LIST_START = exports.LIST_START = Symbol('listStart');
var LIST_END = exports.LIST_END = Symbol('listEnd');
var LIST_CON = exports.LIST_CON = Symbol('listCon');
var ABBERVIATION = exports.ABBERVIATION = Symbol('abberviation');

// Abberiviations.
var QUOTE = exports.QUOTE = Symbol('quote');
var QUASIQUOTE = exports.QUASIQUOTE = Symbol('quasiquote');
var UNQUOTE = exports.UNQUOTE = Symbol('unquote');
var UNQUOTE_SPLICING = exports.UNQUOTE_SPLICING = Symbol('unquote-splicing');
var SYNTAX = exports.SYNTAX = Symbol('syntax');
var QUASISYNTAX = exports.QUASISYNTAX = Symbol('quasisyntax');
var UNSYNTAX = exports.UNSYNTAX = Symbol('unsyntax');
var UNSYNTAX_SPLICING = exports.UNSYNTAX_SPLICING = Symbol('unsyntax-splicing');

// Used for list brackets.
var ROUND_BRACKET = exports.ROUND_BRACKET = Symbol('roundBracket');
var SQUARE_BRACKET = exports.SQUARE_BRACKET = Symbol('squareBracket');

var DEFAULT_STATE = 0;
var COMMENT_BLOCK_STATE = 1;

var CHAR_MAPPING = {
  nul: '\0',
  alarm: '\x07',
  backspace: '\b',
  tab: '\t',
  linefeed: '\n',
  vtab: '\v',
  page: '\x0c',
  return: '\r',
  esc: '\b',
  space: ' ',
  delete: '\xff' // Is this right?
};

var SYNTAX_TABLE = [
// default state
[[/;(.+)$|(#!r(5|6|7)rs)/gm, function () {
  return { type: COMMENT };
}], [/#;/g, function () {
  return { type: COMMENT_IGNORE };
}], [/#\|/g, function (machine) {
  machine.state = COMMENT_BLOCK_STATE;
  machine.depth = 0;
  return { type: COMMENT_BLOCK_START };
}],
// [/((#[bBoOdDxX])?(#[iIeE])?|(#[iIeE])?(#[bBoOdDxX])?)?(\+|-)?/],
// Support only decimals for now...
[/((\+|-)?)([0-9]+(\.[0-9]+)?(e-?[0-9]+)?|nan\.0|inf\.0)/g, function (_, v) {
  var value = void 0;
  switch (v[0]) {
    case '+inf.0':
    case 'inf.0':
      value = Infinity;
      break;
    case '-inf.0':
      value = -Infinity;
      break;
    case 'nan.0':
    case '+nan.0':
    case '-nan.0':
      value = NaN;
      break;
    default:
      value = parseFloat(v[0]);
  }
  return {
    type: NUMBER,
    value: value
  };
}],
// [/[a-zA-Z!$%&*/:<=>?\^_~+\-]/]
// peculiar identifier is not implemented yet
[/((?:[\0-\x08\x0E-\x1F!\$-&\*\+\-/:<-Z\\\^_a-\x9F\xA1-\u167F\u1681-\u180D\u180F-\u1FFF\u200B-\u2027\u202A-\u202E\u2030-\u205E\u2060-\u2FFF\u3001-\uD7FF\uE000-\uFEFE\uFF00-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])|\\x[0-9A-Fa-f]+)((?:[\0-\x08\x0E-\x1F!\$-&\*-:<-Z\\\^_a-\x9F\xA1-\u167F\u1681-\u180D\u180F-\u1FFF\u200B-\u2027\u202A-\u202E\u2030-\u205E\u2060-\u2FFF\u3001-\uD7FF\uE000-\uFEFE\uFF00-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF]))*|\.\.\./g, function (_, v) {
  return {
    type: IDENTIFIER,
    value: v[0]
  };
}], [new RegExp('#\\\\(x([0-9a-fA-F]+)|nul|alarm|backspace|tab|linefeed' + '|newline|vtab|page|return|esc|space|delete|.)', 'gu'), function (_, v) {
  return {
    type: CHARACTER,
    value: CHAR_MAPPING[v[1]] || v[1].replace(/x([0-9a-fA-F]+)/g, function (match, p1) {
      return String.fromCodePoint(parseInt(p1, 16));
    })
  };
}], [/"(((?:[\0-!#-\[\]-\uD7FF\uE000-\uFFFF]|[\uD800-\uDBFF][\uDC00-\uDFFF]|[\uD800-\uDBFF](?![\uDC00-\uDFFF])|(?:[^\uD800-\uDBFF]|^)[\uDC00-\uDFFF])|\\["\\abfnrtv]|\\[\t-\r \xA0\u1680\u180E\u2000-\u200A\u2028\u2029\u202F\u205F\u3000\uFEFF]*\n[\t-\r \xA0\u1680\u180E\u2000-\u200A\u2028\u2029\u202F\u205F\u3000\uFEFF]*|\\x([0-9A-Fa-f]+))*)"/g, function (_, v) {
  return {
    type: STRING,
    value: v[1].replace(/\\n/g, '\n').replace(/\\\\/g, '\\').replace(/\\r/g, '\r').replace(/\\b/g, '\b').replace(/\\t/g, '\t').replace(/\\v/g, '\v').replace(/\\"/g, '\"').replace(/\\f/g, '\f')
    // Assume \a (alarm) is \x07
    .replace(/\\a/g, '\x07').replace(/\\\s*\n\s*/g, '').replace(/\\x([0-9a-fA-F]+)/g, function (match, p1) {
      return String.fromCodePoint(parseInt(p1, 16));
    })
  };
}], [/#(t|T|f|F)/g, function (_, v) {
  return {
    type: BOOLEAN,
    value: v[0] === '#t' || v[0] === '#T'
  };
}], [/#vu8\(/g, function () {
  return { type: BYTE_VECTOR_START, value: ROUND_BRACKET };
}], [/#\(/g, function () {
  return { type: VECTOR_START, value: ROUND_BRACKET };
}], [/\(/g, function () {
  return { type: LIST_START, value: ROUND_BRACKET };
}], [/\[/g, function () {
  return { type: LIST_START, value: SQUARE_BRACKET };
}], [/\)/g, function () {
  return { type: LIST_END, value: ROUND_BRACKET };
}], [/\]/g, function () {
  return { type: LIST_END, value: SQUARE_BRACKET };
}], [/#'/g, function () {
  return { type: ABBERVIATION, value: SYNTAX };
}], [/#`/g, function () {
  return { type: ABBERVIATION, value: QUASISYNTAX };
}], [/#,@/g, function () {
  return { type: ABBERVIATION, value: UNSYNTAX_SPLICING };
}], [/#,/g, function () {
  return { type: ABBERVIATION, value: UNSYNTAX };
}], [/'/g, function () {
  return { type: ABBERVIATION, value: QUOTE };
}], [/`/g, function () {
  return { type: ABBERVIATION, value: QUASIQUOTE };
}], [/,@/g, function () {
  return { type: ABBERVIATION, value: UNQUOTE_SPLICING };
}], [/,/g, function () {
  return { type: ABBERVIATION, value: UNQUOTE };
}], [/\.(?=(^|[()\[\]";#\s]))/g, function () {
  return { type: LIST_CON };
}], [/\s+/g, function () {
  return undefined;
}]
// TODO: Implement , ,@ #' #` #, #,@
], [// command block state
[/#\|/g, function (machine) {
  machine.depth++;
  return { type: COMMENT_BLOCK_START };
}], [/\|#/g, function (machine) {
  if (machine.depth === 0) {
    machine.state = DEFAULT_STATE;
  } else {
    machine.depth--;
  }
  return { type: COMMENT_BLOCK_END };
}], [/([^\|#]+|.)/g, function () {
  return undefined;
}]]];

function tokenize(code) {
  // Memorize line break position
  var lineBreaks = [],
      result = void 0,
      lineBreakRegex = /\n|\r/g;
  // This is awful???
  result = lineBreakRegex.exec(code);
  while (result != null) {
    lineBreaks.push(result.index);
    result = lineBreakRegex.exec(code);
  }
  var line = 1;
  var lineStart = 0;
  var machine = {
    state: 0
  };
  var output = [];
  var index = 0;
  while (index < code.length) {
    // Cycle through all syntaxes...
    var syntaxes = SYNTAX_TABLE[machine.state];
    var results = [];
    for (var i = 0; i < syntaxes.length; ++i) {
      var syntax = syntaxes[i];
      var pattern = syntax[0];
      pattern.lastIndex = index;
      var _result = pattern.exec(code);
      if (!_result || _result.index !== index) continue;
      var callback = syntax[1];
      results.push([_result, callback]);
    }
    // results.sort((a, b) => a[0].length - b[0].length);
    if (results.length === 0) {
      var sliced = code.slice(index);
      var pos = sliced.search(/\s/);
      if (pos === -1) pos = sliced.length;
      while (index > lineBreaks[0]) {
        lineStart = lineBreaks.shift();
        line += 1;
      }
      var error = new Error('Unknown token ' + sliced.slice(0, pos));
      error.line = line;
      error.column = index - lineStart + 1;
      throw error;
    }
    var next = false;
    for (var _i = 0; _i < results.length; ++_i) {
      var _result2 = results[_i][1](machine, results[_i][0]);
      if (_result2 !== undefined) {
        // Inject token position...
        while (index > lineBreaks[0]) {
          lineStart = lineBreaks.shift();
          line += 1;
        }
        _result2.line = line;
        _result2.column = index - lineStart + 1;
        output.push(_result2);
        index = results[_i][0][0].length + index;
        next = true;
        break;
      }
    }
    if (!next) {
      index = results[results.length - 1][0][0].length + index;
    }
  }
  return output;
}